@@ -1,5 +1,5 @@
 ##
-## $Id: SQLiteSupport.R,v 1.2 2002/09/05 02:39:38 dj Exp dj $
+## $Id: SQLiteSupport.R 185 2006-09-22 20:39:54Z sethf $
 ##
 ## Copyright (C) 1999-2002 The Omega Project for Statistical Computing.
 ##
@@ -19,7 +19,7 @@
 ##
 
 "sqliteInitDriver" <- 
-function(max.con = 16, fetch.default.rec = 500, force.reload=F)
+function(max.con = 16, fetch.default.rec = 500, force.reload=FALSE)
 ## return a manager id
 {
   config.params <- as.integer(c(max.con, fetch.default.rec))
@@ -89,13 +89,30 @@ function(obj, what="", ...)
 ## the SQLite API as the default database (SQLite config specific)
 ## while NULL means "no database".
 "sqliteNewConnection"<- 
-function(drv, dbname = "", mode=0)
+function(drv, dbname = "", mode=0, cache_size=NULL, synchronous=0)
 {
+  if (!is.null(dbname)) 
+    dbname <- path.expand(dbname)
   con.params <- as.character(c(dbname, mode))
   drvId <- as(drv, "integer")
   conId <- .Call("RS_SQLite_newConnection", drvId, con.params, 
                  PACKAGE ="RSQLite")
-  new("SQLiteConnection", Id = conId)
+  con <- new("SQLiteConnection", Id = conId)
+
+  ## experimental PRAGMAs
+  try({
+    if(!is.null(cache_size))
+      dbGetQuery(con, sprintf("PRAGMA cache_size=%d", as.integer(cache_size)))
+    if(is.numeric(synchronous))
+      nsync <- as.integer(synchronous)
+    else if(is.character(synchronous))
+      nsync <- 
+        pmatch(tolower(synchronous), c("off", "normal", "full"), nomatch = 1) - 1
+    else nsync <- 0
+    dbGetQuery(con, sprintf("PRAGMA synchronous=%d", as.integer(nsync)))
+   }, silent = TRUE)
+
+  con
 }
 
 "sqliteDescribeConnection" <- 
@@ -181,7 +198,7 @@ function(con, statement, limit = -1)
 ## helper function: it exec's *and* retrieves a statement. It should
 ## be named somehting else.
 "sqliteQuickSQL" <- 
-function(con, statement)
+function(con, statement, ...)
 {
    nr <- length(dbListResults(con))
    if(nr>0){                     ## are there resultSets pending on con?
@@ -194,7 +211,7 @@ function(con, statement)
       invisible()
       return(NULL)
    }
-   res <- sqliteFetch(rs, n = -1)
+   res <- sqliteFetch(rs, n = -1, ...)
    if(dbHasCompleted(rs))
       dbClearResult(rs)
    else 
@@ -314,7 +331,7 @@ function(con, name, ...)
 }
 ## this is exactly the same as ROracle's oraReadTable
 "sqliteReadTable" <-  
-function(con, name, row.names = "row.names", check.names = TRUE, ...)
+function(con, name, row.names = "row_names", check.names = TRUE, ...)
 ## Should we also allow row.names to be a character vector (as in read.table)?
 ## is it "correct" to set the row.names of output data.frame?
 ## Use NULL, "", or 0 as row.names to prevent using any field as row.names.
@@ -327,16 +344,16 @@ function(con, name, row.names = "row.names", check.names = TRUE, ...)
    ## should we set the row.names of the output data.frame?
    nms <- names(out)
    j <- switch(mode(row.names),
-           "character" = if(row.names=="") 0 else
+           character = if(row.names=="") 0 else
                match(tolower(row.names), tolower(nms), 
                      nomatch = if(missing(row.names)) 0 else -1),
-           "numeric", "logical" = row.names,
-           "NULL" = 0,
+           numeric=, logical = row.names,
+           NULL = 0,
            0)
    if(as.numeric(j)==0) 
       return(out)
    if(is.logical(j)) ## Must be TRUE
-      j <- match("row.names", tolower(nms), nomatch=0) 
+      j <- match(row.names, tolower(nms), nomatch=0) 
    if(j<1 || j>ncol(out)){
       warning("row.names not set on output data.frame (non-existing field)")
       return(out)
@@ -349,44 +366,46 @@ function(con, name, row.names = "row.names", check.names = TRUE, ...)
    out
 }
 
-"sqliteWriteTable" <-
-function(con, name, value, field.types, row.names = TRUE, 
-  overwrite=FALSE, append=FALSE, ...)
-## TODO: This function should execute its sql as a single transaction,
-## and allow converter functions.
-## Create table "name" (must be an SQL identifier) and populate
-## it with the values of the data.frame "value"
-## BUG: In the unlikely event that value has a field called "row.names"
-## we could inadvertently overwrite it (here the user should set row.names=F)
-## (I'm reluctantly adding the code re: row.names -- I'm not 100% comfortable
-## using data.frames as the basic data for relations.)
+"dbBuildTableDefinition" <-
+function(dbObj, name, value, field.types = NULL, row.names = TRUE, ...)
 {
-  if(overwrite && append)
-    stop("overwrite and append cannot both be TRUE")
   if(!is.data.frame(value))
     value <- as.data.frame(value)
-  if(row.names){
+  if(!is.null(row.names) && row.names){
     value  <- cbind(row.names(value), value)  ## can't use row.names= here
     names(value)[1] <- "row.names" 
   }
-  if(missing(field.types) || is.null(field.types)){
+  if(is.null(field.types)){
     ## the following mapping should be coming from some kind of table
     ## also, need to use converter functions (for dates, etc.)
-    field.types <- sapply(value, dbDataType, dbObj = con)
+    field.types <- sapply(value, dbDataType, dbObj = dbObj)
   } 
   i <- match("row.names", names(field.types), nomatch=0)
   if(i>0) ## did we add a row.names value?  If so, it's a text field.
-    field.types[i] <- dbDataType(con, field.types$row.names)
-  names(field.types) <- make.db.names(con, names(field.types), allow.keywords=F)
+    field.types[i] <- dbDataType(dbObj, field.types$row.names)
+  names(field.types) <- 
+    make.db.names(dbObj, names(field.types), allow.keywords = FALSE)
+
+  ## need to create a new (empty) table
+  flds <- paste(names(field.types), field.types)
+  paste("CREATE TABLE", name, "\n(", paste(flds, collapse=",\n\t"), "\n)")
+}
+
+"sqliteImportFile" <-
+function(con, name, value, field.types = NULL, overwrite = FALSE, 
+  append = FALSE, header, row.names, nrows = 50, sep = ",", 
+  eol="\n", skip = 0, ...)
+{
+  if(overwrite && append)
+    stop("overwrite and append cannot both be TRUE")
 
   ## Do we need to clone the connection (ie., if it is in use)?
   if(length(dbListResults(con))!=0){ 
     new.con <- dbConnect(con)              ## there's pending work, so clone
     on.exit(dbDisconnect(new.con))
   } 
-  else {
+  else 
     new.con <- con
-  }
 
   if(dbExistsTable(con,name)){
     if(overwrite){
@@ -399,41 +418,77 @@ function(con, name, value, field.types, row.names = TRUE,
       warning(paste("table", name, "exists in database: aborting dbWriteTable"))
       return(FALSE)
     }
-  } 
-  if(!dbExistsTable(con,name)){      ## need to re-test table for existance 
-    ## need to create a new (empty) table
-    sql1 <- paste("create table ", name, "\n(\n\t", sep="")
-    sql2 <- paste(paste(names(field.types), field.types), collapse=",\n\t", 
-                  sep="")
-    sql3 <- "\n)\n"
-    sql <- paste(sql1, sql2, sql3, sep="")
+  }
+
+  ## compute full path name (have R expand ~, etc)
+  fn <- file.path(dirname(value), basename(value))
+  if(missing(header) || missing(row.names)){
+    f <- file(fn, open="r")
+    if(skip>0) 
+      readLines(f, n=skip)
+    txtcon <- textConnection(readLines(f, n=2))
+    flds <- count.fields(txtcon, sep)
+    close(txtcon)
+    close(f)
+    nf <- length(unique(flds))
+  }
+  if(missing(header)){
+    header <- nf==2
+  }
+  if(missing(row.names)){
+    if(header)
+      row.names <- if(nf==2) TRUE else FALSE
+    else
+      row.names <- FALSE
+  }
+
+  new.table <- !dbExistsTable(con, name)
+  if(new.table){
+    ## need to init table, say, with the first nrows lines
+    d <- read.table(fn, sep=sep, header=header, skip=skip, nrows=nrows, ...)
+    sql <- 
+      dbBuildTableDefinition(new.con, name, d, field.types = field.types,
+        row.names = row.names)
     rs <- try(dbSendQuery(new.con, sql))
     if(inherits(rs, ErrorClass)){
-      warning("could not create table: aborting assignTable")
+      warning("could not create table: aborting sqliteImportFile")
       return(FALSE)
     } 
     else 
       dbClearResult(rs)
   }
-
-  fn <- tempfile("rsdbi")
-  dots <- list(...)
-  safe.write(value, file = fn, batch = dots$batch)
-  on.exit(unlink(fn), add = TRUE)
-  sql4 <- paste("COPY '", name,"' FROM '", fn, "' USING DELIMITERS ','", sep="")
-  rs <- try(dbSendQuery(new.con, sql4))
-  if(inherits(rs, ErrorClass)){
-    warning("could not load data into table")
+  else if(!append){
+    warning(sprintf("table %s already exists -- use append=TRUE?", name))
+  }
+  rc <- 
+      try({
+         skip <- skip + as.integer(header)
+         conId <- as(con, "integer")
+         .Call("RS_SQLite_importFile", conId, name, fn, sep, eol,
+            as(skip, "integer"), PACKAGE = "RSQLite")
+      })
+  if(inherits(rc, ErrorClass)){
+    if(new.table) dbRemoveTable(new.con, name)
     return(FALSE)
-  } 
-  else 
-    dbClearResult(rs)
+  }
   TRUE
 }
+
+"sqliteWriteTable" <-
+function(con, name, value, row.names = TRUE, ...)
+{
+  fn <- tempfile("rsdbi")
+  on.exit(unlink(fn), add = TRUE)
+  dots <- list(...)
+  safe.write(value, file = fn, batch = dots$batch, row.names = row.names)
+  a <- list(con = con, name = name, value = fn, header = TRUE)
+  do.call("sqliteImportFile", c(a, dots))
+}
  
 ## from ROracle, except we don't quote strings here.
 "safe.write" <- 
-function(value, file, batch, ..., quote.string = FALSE)
+function(value, file, batch, row.names = TRUE, ..., 
+  sep = ',', eol = '\n', quote.string = FALSE)
 ## safe.write makes sure write.table don't exceed available memory by batching
 ## at most batch rows (but it is still slowww)
 {  
@@ -449,14 +504,10 @@ function(value, file, batch, ..., quote.string = FALSE)
    from <- 1 
    to <- min(batch, N)
    while(from<=N){
-      if(usingR())
-         write.table(value[from:to, drop=FALSE], file = file, append = TRUE, 
-               quote = quote.string, sep=",", na = .SQLite.NA.string, 
-               row.names=FALSE, col.names=FALSE, eol = '\n', ...)
-      else
-         write.table(value[from:to, drop=FALSE], file = file, append = TRUE, 
-               quote.string = quote.string, sep=",", na = .SQLite.NA.string, 
-               dimnames.write=FALSE, end.of.row = '\n', ...)
+     write.table(value[from:to,, drop=FALSE], file = file, 
+        append = from>1,
+        quote = quote.string, sep=sep, na = .SQLite.NA.string, 
+        row.names=row.names, col.names=(from==1), eol = eol, ...)
       from <- to+1
       to <- min(to+batch, N)
    }
@@ -471,7 +522,7 @@ function(value, file, batch, ..., quote.string = FALSE)
   rs.class <- data.class(obj)
   rs.mode <- storage.mode(obj)
   if(rs.class=="numeric"){
-    sql.type <- if(rs.mode=="integer") "bigint" else  "double"
+    sql.type <- if(rs.mode=="integer") "int" else  "double"
   } 
   else {
     sql.type <- switch(rs.class,
